from dataverk import utils
from dataverk.dataverk import Dataverk
from dataverk.dataverk import Dataverk as Client
from dataverk.datapackage import Datapackage

__all__ = ['connectors',
           'utils',
           'Datapackage',
           'Client',
           'Dataverk'
           ]



from collections import Mapping
from enum import Enum
from urllib3.util import parse_url
from dataverk.connectors.databases.base import DBBaseConnector
from dataverk.connectors import (
    OracleConnector,
    PostgresConnector,
    Db2Connector,
    SqliteConnector,
)
from dataverk.exceptions import dataverk_exceptions


class DbType(Enum):
    ORACLE = "oracle"
    POSTGRES = "postgres"
    DB2 = "ibm_db"
    SQLITE = "sqlite"


def get_db_connector(settings_store: Mapping, source: str) -> DBBaseConnector:
    """ Factory function returning connector type

    :param settings_store: Mapping object with project specific configurations
    :param source: str
    :return: database connector
    """
    try:
        connection_string = parse_url(settings_store["db_connection_strings"][source])
    except KeyError:
        raise dataverk_exceptions.IncompleteSettingsObject(
            f"Database connection string not found in settings file. "
            f"Unable to establish connection to database: {source}"
        )

    if DbType.ORACLE.value in connection_string.scheme.lower():
        return OracleConnector(settings_store=settings_store, source=source)
    elif DbType.POSTGRES.value in connection_string.scheme.lower():
        return PostgresConnector(settings_store=settings_store, source=source)
    elif DbType.DB2.value in connection_string.scheme.lower():
        return Db2Connector(settings_store=settings_store, source=source)
    elif DbType.SQLITE.value in connection_string.scheme.lower():
        return SqliteConnector(settings_store=settings_store, source=source)
    else:
        raise NotImplementedError(
            f"""Database type {connection_string.scheme.lower()} is not supported.
            Supported types are {[name.value for name in DbType]}"""
        )



from urllib3.util import parse_url
from collections.abc import Mapping
from dataverk.connectors.databases.base import DBBaseConnector


class OracleConnector(DBBaseConnector):
    def __init__(self, settings_store: Mapping, source: str):
        super().__init__(settings_store, source)

    def _connection_string(self):
        connection_string = self.settings["db_connection_strings"][self.source]
        return OracleConnector._format_connection_string(connection_string)

    @staticmethod
    def _format_connection_string(connection_string):
        parsed_url = parse_url(connection_string)
        if "?" not in parsed_url.request_uri:
            return str(parsed_url).replace(
                parsed_url.request_uri,
                f"/?service_name={parsed_url.request_uri.split('/')[1]}",
            )
        else:
            return connection_string



import json
from enum import Enum
from io import BytesIO
from collections.abc import Mapping
from google.cloud import storage
from google.cloud import exceptions as gcloud_exceptions
from google.oauth2 import service_account
from pathlib import Path
from dataverk.connectors.storage.utils import blob_metadata
from dataverk.exceptions import dataverk_exceptions
from dataverk.connectors.storage.bucket_storage_base import BucketStorageBase


class BlobMetadataFormat(str, Enum):
    MARKDOWN: str = "markdown"
    OBJECT: str = "object"


class GoogleStorageConnector(BucketStorageBase):
    """Google Storage connector"""

    def __init__(self, bucket_name: str, settings: Mapping):
        super().__init__()
        self.bucket = self._get_bucket_connection(bucket_name, settings)

    def write(
        self, data: str, destination_blob_name: str, fmt: str, metadata: dict = None
    ):
        """Write string to a bucket."""
        try:
            name = f"{destination_blob_name}.{fmt}"
            blob = self.bucket.blob(name)
            self._set_blob_metadata(blob=blob, fmt=fmt, metadata=metadata)
            blob.upload_from_string(data)
        except gcloud_exceptions.GoogleCloudError as error:
            self.log.error(f"Error writing file {name} to google storage: {error}")
            raise gcloud_exceptions.GoogleCloudError(error)
        else:
            self.log.info(
                f"{destination_blob_name} successfully written to {blob.public_url}"
            )
            return f"{blob.public_url}"

    def read(self, blob_name: str, **kwargs):
        """Downloads a blob from the bucket."""
        blob = self.bucket.blob(blob_name)
        string_buffer = BytesIO()
        try:
            blob.download_to_file(string_buffer)
        except gcloud_exceptions.GoogleCloudError as error:
            self.log.error(f"Unable to read blob {blob_name}: {error}")
            raise gcloud_exceptions.GoogleCloudError(error)
        else:
            self.log.info(f"Blob {blob_name} read to string to string")
            return string_buffer.getvalue().decode()

    def get_blob_metadata(self, blob_name: str, metadata_format: str = "markdown"):
        """Prints out a blob's metadata."""
        blob = self.bucket.get_blob(blob_name)

        if metadata_format == BlobMetadataFormat.MARKDOWN.value:
            return blob_metadata.as_markdown(blob)
        elif metadata_format == BlobMetadataFormat.OBJECT.value:
            return blob_metadata.as_json(blob)
        else:
            raise NotImplementedError(
                f"Blob metadata format {metadata_format} is not supported. "
                f"Supported formats are: {[name.value for name in BlobMetadataFormat]}"
            )

    def _get_bucket_connection(self, bucket_name, settings):
        try:
            storage_client = storage.Client(
                credentials=self._get_gcp_credentials(settings)
            )
            return self._get_bucket(storage_client, bucket_name)
        except gcloud_exceptions.GoogleCloudError as error:
            self.log.error(f"Unable to get bucket {bucket_name}: {error}")
            raise gcloud_exceptions.GoogleCloudError(error)

    def _get_gcp_credentials(self, settings):
        try:
            info = settings["bucket_storage"]["gcs"]["credentials"]
        except KeyError as missing:
            raise dataverk_exceptions.IncompleteSettingsObject(f"{missing}")
        else:
            info = self.parse_gcp_credentials(info)
            scope = "https://www.googleapis.com/auth/cloud-platform"
            credentials = service_account.Credentials.from_service_account_info(
                info, scopes=(scope,)
            )
            return credentials

    def _get_bucket(self, storage_client, bucket_name):
        try:
            bucket = storage_client.get_bucket(bucket_name)
            bucket.acl.reload()
            return bucket
        except gcloud_exceptions.NotFound:
            raise dataverk_exceptions.StorageBucketDoesNotExist(bucket_name)

    def _set_blob_metadata(self, blob, fmt: str, metadata: Mapping):
        if fmt.lower().endswith("gz"):
            blob.content_encoding = f"gzip"
        blob.cache_control = "no-cache"
        blob.metadata = metadata

    @staticmethod
    def parse_gcp_credentials(info):
        if isinstance(info, str):
            if GoogleStorageConnector.is_json_file(info):
                with open(Path(info)) as file:
                    return json.load(file)
            else:
                return json.loads(info)
        else:
            return info

    @staticmethod
    def is_json_file(info: str):
        return info.endswith(".json")



import math
import pandas as pd

from collections.abc import Sequence
from dataverk.abc.base import DataverkBase
from dataverk.utils import dataverk_doc_address
from dataverk.exceptions import dataverk_exceptions
from dataverk.datapackage import Datapackage
from dataverk.context import EnvStore
from dataverk.dataverk_context import DataverkContext
from dataverk.connectors import KafkaConnector, kafka, JSONStatConnector
from dataverk.connectors.databases import db_connector_factory
from dataverk.elastic_search_updater import ElasticSearchUpdater
from dataverk.connectors.elasticsearch import ElasticsearchConnector
from dataverk.package_publisher import PackagePublisher
from dataverk.utils.anonymization import anonymize_replace


class Dataverk(DataverkBase):

    def __init__(self, resource_path: str=".", env_file: str='.env', auth_token: str=None):
        super().__init__()
        env_store = EnvStore.safe_create(env_file)
        self._context = DataverkContext(env_store, resource_path, auth_token)

    @property
    def context(self):
        return self._context

    def read_sql(self, source: str, sql: str, connector: str='Oracle', verbose_output: bool=False, *args, **kwargs) -> pd.DataFrame:
        """ Read pandas dataframe from SQL database

        :param source: str: database source reference
        :param sql: str: sql query or file with sql query
        :param connector: str: Database connector (default oracle)
        :param verbose_output: bool: flag for verbose output option
        :return: pd.Dataframe: Dataframe with result
        """
        conn = db_connector_factory.get_db_connector(settings_store=self.context.settings, source=source)
        query = self._get_sql_query(sql=sql)

        with conn:
            return conn.get_pandas_df(query=query, verbose_output=verbose_output, *args, **kwargs)

    def read_kafka_message_fields(self, topics: Sequence, fetch_mode: str = "from_beginning") -> pd.DataFrame:
        """ Read single kafka message from topic and return list of message fields

        :param topics: Sequence of topics to subscribe to
        :param fetch_mode: str describing fetch mode (from_beginning, last_committed_offset), default last_committed_offset
        :return: list: fields in kafka message
        """
        consumer = kafka.get_kafka_consumer(settings=self.context.settings, topics=topics, fetch_mode=fetch_mode)
        conn = KafkaConnector(consumer=consumer, settings=self.context.settings, topics=topics, fetch_mode=fetch_mode)

        return conn.get_message_fields()

    def read_kafka(self, topics: Sequence, strategy=None, fields=None, fetch_mode: str = "from_beginning", max_mesgs: int=math.inf) -> pd.DataFrame:
        """ Read kafka topics and return pandas dataframe

        :param strategy: function or lambda passed to the kafka consumer for aggregating data on the fly
        :param fields: requested fields in kafka message
        :param max_mesgs: max number of kafka messages to read
        :param topics: Sequence of topics to subscribe to
        :param fetch_mode: str describing fetch mode (from_beginning, last_committed_offset), default last_committed_offset
        :return: pandas.Dataframe
        """
        consumer = kafka.get_kafka_consumer(settings=self.context.settings, topics=topics, fetch_mode=fetch_mode)
        conn = KafkaConnector(consumer=consumer, settings=self.context.settings, topics=topics, fetch_mode=fetch_mode)

        return conn.get_pandas_df(strategy=strategy, fields=fields, max_mesgs=max_mesgs)

    def read_json_stat(self, url, params=None):
        """ Read json-stat return pandas dataframe

        :param url: str: path to resource
        :param params: optional request parameters
        :return: pandas.Dataframe
        """
        conn = JSONStatConnector()
        return conn.get_pandas_df(url, params=params)

    def anonymize(self, df, eval_column, anonymize_columns=None, evaluator=lambda x: x < 4, replace_by="*",
                  anonymize_eval=True):
        """ Replace values in columns when condition in evaluator is satisfied

        :param df: pandas DataFrame
        :param eval_column: name of column to evaluate for anonymization
        :param anonymize_columns: optional, column name or list of column(s) to anonymize if value in eval_column satisfies the
        condition given in evaluator, default=None
        :param evaluator: lambda function, condition for anonymization based on values in eval_column, default=lambda x: x < 4
        :param replace_by: value or list or dict of values to replace by. List or dict passed must have same length as the number
        of columns to anonymize. Elements in list passed should in addition have the same order as columns in

        a) anonymize_columns + eval_columns if anonymize_eval=True and eval_column is _not_ given in anonymize_columns
        b) anonymize_columns                if anonymize_eval=True and eval_column is given in anonymize_columns
                                            or anonymize_eval=False
        c) eval_column                      if anonymize_eval=True and anonymize_columns is None or anonymize_columns=[]

        The order of values to replace by in dictionary does not matter.

        :param anonymize_eval, bool, whether to anonymize eval_column, default=True

        :return: anonymized pandas DataFrame
        """
        return anonymize_replace(df=df, eval_column=eval_column, anonymize_columns=anonymize_columns,
                                 evaluator=evaluator, replace_by=replace_by,
                                 anonymize_eval=anonymize_eval)

    def to_sql(self, df: pd.DataFrame, table: str, sink: str, connector: str='Oracle', if_exists: str='append', *args, **kwargs):
        """ Write records in dataframe to a SQL database table

        :param df: pd.Dataframe: Dataframe to write
        :param table: str: Table in db to write to
        :param sink: str: target database name reference
        :param connector: str: Connector type (default: Oracle)
        :param if_exists: str: Action if table already exists in database (default: replace)
        """
        conn = db_connector_factory.get_db_connector(settings_store=self._context.settings, source=sink)

        with conn:
            return conn.persist_pandas_df(table, df=df, if_exists=if_exists, *args, **kwargs)

    def publish(self, datapackage: Datapackage):
        resources = datapackage.resources
        metadata = datapackage.datapackage_metadata

        # Publish resources to buckets
        package_publisher = PackagePublisher(datapackage_metadata=metadata, settings_store=self._context.settings, env_store={})
        package_publisher.publish(resources=resources)

        # Publish metadata to elastic search
        try:
            es_conn = ElasticsearchConnector(self._context.settings)
        except dataverk_exceptions.IncompleteSettingsObject:
            self.log.warning(
                f"""Could not publish metadata to elastic search index.
                ES index not specified in settings object.
                See {dataverk_doc_address} for guidelines on how to setup the settings file."""
            )
        else:
            eu = ElasticSearchUpdater(es_conn, metadata)
            eu.publish()

    def _get_sql_query(self, sql):
        if self._is_sql_file(sql):
            return self.context.get_sql_query(sql)
        else:
            return sql

    @staticmethod
    def _is_sql_file(source):
        return '.sql' in source



import logging


class Logger:

    def __init__(self, user: str, connector_name: str):
        self._logger = logging.getLogger(connector_name)
        self._configure_logger(user)

    @property
    def log(self):
        return self._logger

    def _configure_logger(self, user) -> None:
        self._logger.propagate = False
        self._logger.setLevel(logging.INFO)

        if not len(self._logger.handlers):
            handler = self._create_handler(user)
            self._logger.addHandler(handler)

    def _create_handler(self, user) -> logging.Handler:
        handler = logging.StreamHandler()
        formatter = self._create_formatter(user)
        handler.setFormatter(formatter)
        return handler

    def _create_formatter(self, user) -> logging.Formatter:
        return logging.Formatter(f"%(levelname)s:{user}:%(name)s:%(asctime)s: %(message)s",
                                 "%Y-%m-%d %H:%M:%S")



from dataverk.abc.base import DataverkBase


class BucketStorageBase(DataverkBase):
    def __init__(self):
        super().__init__()

    def write(self, data, destination_blob_name: str, fmt: str, **kwargs):
        raise NotImplementedError(
            f"Abstract method. Needs to be implemented in subclass"
        )

    def read(self, blob_name: str, **kwargs):
        raise NotImplementedError(
            f"Abstract method. Needs to be implemented in subclass"
        )



import urllib3

from collections import Mapping
from dataverk.abc.base import DataverkBase
from dataverk.connectors.elasticsearch import ElasticsearchConnector
from datetime import datetime


class ElasticSearchUpdater(DataverkBase):

    def __init__(self, es_index: ElasticsearchConnector, datapackage_metadata: Mapping):
        super().__init__()
        self._es_index = es_index
        self.datapackage_json = datapackage_metadata

    def publish(self):
        """ Updates ES index with metadata for the datapackage

        :return: None
        """

        try:
            dp = self.datapackage_json
            id = dp["id"]
            title = dp.get('title', 'title missing')
            desc = dp.get('description', 'description missing')
            js = {
                "id": id,
                "versionInfo": dp.get("versionInfo", "0.0.1"),
                "versionNotes": dp.get("versionNotes", []),
                "type": dp.get("type", ""),
                "format": dp.get("format", ""),
                "suggest": title + ' ' + desc,
                "description": desc,
                "title": title,
                "license": dp.get("license", {'name': 'CC BY 4.0', 'url': 'http://creativecommons.org/licenses/by/4.0/deed.no'}),
                "language": dp.get("language", "Norsk"),
                "accrualPeriodicity": dp.get("accrualPeriodicity", ""),
                "temporal": dp.get("temporal", ""),
                "category": dp.get("category", ""),
                "status": dp.get("status", ""),
                "rights": dp.get("rights", []),
                "byteSize": dp.get("byteSize", []),
                "provenance": dp.get("provenance", ""),
                "issued": dp.get("issued", datetime.now().isoformat()),
                "modified": dp.get("modified", datetime.now().isoformat()),
                "distribution": dp.get("distribution", []),
                "keyword": dp.get("keywords", []),
                "term": dp.get("term", []),
                "theme": dp.get("theme", []),
                "accessRights": dp.get("accessRights", ['non-public']),
                "accessRightsComment": dp.get("accessRightsComment", ""),
                "publisher": dp.get("publisher", {'name': 'Arbeids- og velferdsetaten (NAV)', 'publisher_url': 'https://www.nav.no'}),
                "creator": dp.get("creator", {'name': 'NAV kunnskapsavdelingen', 'email': 'statistikk@nav.no'}),
                "contactPoint": dp.get("contactPoint", []),
                "spatial": dp.get("spatial", []),
                "url": f"{dp.get('path', '')}/datapackage.json",
                "source": dp.get("source", []),
                "sample": dp.get("sample", []),
                "repo": dp.get("repo", ""),
                "notebook": dp.get("notebook", ""),
                "code": dp.get("code", ""),
            }

            resources = dp.get('resources', [])

            resource_names = []
            resource_descriptions = []
            for resource in resources:
                if resource.get('name'):
                    resource_names.append(resource['name'])
                if resource.get('description'):
                    resource_descriptions.append(resource['description'])

            js['resource_names'] = resource_names
            js['resource_descriptions'] = resource_descriptions

            self._es_index.write(id, js)
        except urllib3.exceptions.LocationValueError as err:
            self.log.error(f"write to elastic search failed, host_uri could not be resolved")
            raise urllib3.exceptions.LocationValueError(err)








import os
import errno
import json
import pathlib
import shutil
import stat
import requests
from pathlib import Path
from urllib3.exceptions import LocationParseError
from urllib3.util import url

from dataverk.connectors.storage.google_storage import GoogleStorageConnector


def get_package_resource(resource_name: str, base_path: str, http_headers: dict):
    if os.environ.get("DATAVERK_SETTINGS_PATH", None) is not None:
        return read_file(Path(os.environ.get("DATAVERK_SETTINGS_PATH")).joinpath(resource_name))
    else:
        try:
            parsed_url = url.parse_url(base_path)
        except LocationParseError:
            # parse_url throws exception with local windows paths
            return read_file(Path(base_path).joinpath(resource_name))
        else:
            if parsed_url.scheme == "http" or parsed_url.scheme == "https":
                return requests.get(f"{base_path}/{resource_name}",
                                    headers=http_headers).text
            elif parsed_url.scheme == "gs":
                bucket_conn = GoogleStorageConnector(bucket_name=parsed_url.host)
                return bucket_conn.read(f"{parsed_url.path[1:]}/{resource_name}")
            else:
                return read_file(Path(base_path).joinpath(resource_name))


def write_file(path, content, compressed: bool=False):
    pathlib.Path(path).parent.mkdir(parents=True, exist_ok=True)
    with Path(path).open("w" if not compressed else "wb") as file:
        file.write(content)


def read_file(path, encoding="utf-8"):
    with Path(path).open("r", encoding=encoding) as file:
        return file.read()


def json_to_dict(path: Path):
    file_string = read_file(path)
    return json.loads(file_string)


def remove_whitespace(filename):
    return filename.replace(' ', '_')


def _json_validate_params(file_path: Path):
    if not isinstance(file_path, Path):
        raise TypeError(f"settings_file_path: {file_path} should be a Path object")

    if not file_path.is_file():
        raise FileNotFoundError("The provided url does not resolve to a file")
    if _get_url_suffix(str(file_path)) != "json":
        raise FileNotFoundError("The provided url does not resolve to a json file")


def _get_url_suffix(url: str):
    return url.split(".")[-1]


def win_readonly_directories_force_remove(directory):
    """ Removes directories with read-only files"""
    shutil.rmtree(directory, onerror=_remove_readonly)


def _remove_readonly(func, path, _):
    """Clear the readonly bit and reattempt the removal"""
    os.chmod(path, stat.S_IWRITE)
    func(path)



from dataverk.utils import dataverk_doc_address


class EnvironmentVariableNotSet(Exception):

    def __init__(self, env: str):
        self._env = env

    def __str__(self):
        return f"""
                Required environment variable {self._env} is not set
                """


class IncompleteSettingsObject(Exception):

    def __init__(self, missing: str):
        self._missing = missing

    def __str__(self):
        return f"""{self._missing} is not provided in settings object. 
        See {dataverk_doc_address} for guidelines on how to setup settings file."""


class StorageBucketDoesNotExist(Exception):

    def __init__(self, bucket_name):
        self._bucket_name = bucket_name

    def __str__(self):
        return f"""{self._bucket_name} does not exist"""


import math

import pandas as pd
import json
import struct
import requests
import avro
import avro.schema
import avro.io
import io
import time

from dataverk.utils import mapping_util
from kafka import KafkaConsumer
from collections.abc import Mapping, Sequence
from enum import Enum
from datetime import datetime
from dataverk.abc.base import DataverkBase
import streamz


class KafkaFetchMode(Enum):
    FROM_BEGINNING = "from_beginning"
    LAST_COMMITED_OFFSET = "last_commited_offset"


class KafkaConnector(DataverkBase):

    def __init__(self, consumer: KafkaConsumer, settings: Mapping, topics: Sequence, fetch_mode: str):
        """ Dataverk Kafka consumer class

        :param settings: Mapping object containing project settings
        :param topics: Sequence of topics to subscribe to
        :param fetch_mode: str describing fetch mode (from_beginning, last_committed_offset)
        """
        super().__init__()
        self._topics = topics
        self._fetch_mode = fetch_mode
        self._schema_cache = {}
        self._consumer = consumer
        self.log.info(f"KafkaConsumer created with fetch mode set to '{fetch_mode}'")
        self._read_until_timestamp = self._get_current_timestamp_in_ms()
        self._schema_registry_url = mapping_util.safe_get_nested(settings,
                                                                 keys=("kafka", "schema_registry"),
                                                                 default="http://localhost:8081")

    def get_pandas_df(self, strategy=None, fields=None, max_mesgs=math.inf):
        """ Read kafka topics, commits offset and returns result as pandas dataframe

        :return: pd.Dataframe containing kafka messages read. NB! Commits offset
        """
        if strategy is None:
            records = self._read_kafka_raw(max_mesgs, fields)
        else:
            records = self._read_kafka_accumulated(max_mesgs, strategy)
        try:
            df = pd.DataFrame.from_records(records)
        except ValueError:
            df = pd.DataFrame.from_records(records, index=[0])
        self._commit_offsets()
        self._consumer.close()

        return df

    def get_message_fields(self):
        """ Read single kafka message from topic and return message fields

        :return: list: message fields
        """
        for message in self._consumer:
            try:
                schema_res = self._get_schema_from_registry(message=message)
                schema = schema_res.json()["schema"]
            except (AttributeError, KeyError):
                mesg = json.loads(message.value.decode('utf8'))
            else:
                mesg = self._decode_avro_message(schema=schema, message=message)

            return mesg.keys()

    def _read_kafka_raw(self, max_mesgs, fields):
        start_time = time.time()

        self.log.info(f"Reading kafka topic {self._topics}. Fetch mode {self._fetch_mode}")

        data = list()

        for message in self._consumer:
            mesg = self._parse_kafka_message(message)
            data.append(self._extract_requested_fields(mesg, fields))
            if self._is_requested_messages_read(message, max_mesgs, len(data)):
                break

        self.log.info(f"({len(data)} messages read from kafka topic(s) {self._topics} in {time.time() - start_time} sec. Fetch mode {self._fetch_mode}")

        return data

    def _read_kafka_accumulated(self, max_mesgs, strategy):
        data = {}
        mesg_count = 0
        stream = streamz.Stream()
        acc = stream.accumulate(strategy, start=data)

        for message in self._consumer:
            mesg = self._parse_kafka_message(message)
            mesg_count += 1
            stream.emit(mesg)
            if self._is_requested_messages_read(message, max_mesgs, mesg_count):
                break
        return data

    def _parse_kafka_message(self, message):
        try:
            schema_res = self._get_schema_from_registry(message=message)
            schema = schema_res.json()["schema"]
        except (AttributeError, KeyError):
            return json.loads(message.value.decode('utf8'))
        else:
            return self._decode_avro_message(schema=schema, message=message)

    def _get_schema_from_registry(self, message):
        schema_id = struct.unpack(">L", message.value[1:5])[0]
        if schema_id in self._schema_cache:
            return self._schema_cache[schema_id]
        else:
            schema = requests.get(self._schema_registry_url + '/schemas/ids/' + str(schema_id))
            self._schema_cache[schema_id] = schema
            return schema

    @staticmethod
    def _decode_avro_message(schema, message):
        schema = avro.schema.Parse(schema)
        bytes_reader = io.BytesIO(message.value[5:])
        decoder = avro.io.BinaryDecoder(bytes_reader)
        reader = avro.io.DatumReader(schema)
        return reader.read(decoder)

    @staticmethod
    def _get_current_timestamp_in_ms():
        return int(datetime.now().timestamp() * 1000)

    def _is_requested_messages_read(self, message, max_mesgs, mesgs_read):
        if message.timestamp >= self._read_until_timestamp:
            return True
        elif mesgs_read >= max_mesgs:
            return True
        return False

    def _commit_offsets(self):
        """ Commits the offsets to kafka when the KafkaConsumer object is configured with group_id
        """
        if self._consumer.config["group_id"] is not None:
            self._consumer.commit()

    def _extract_requested_fields(self, mesg, fields):
        if fields is not None:
            return [self._extract_field(mesg, field) for field in fields]
        else:
            return mesg

    @staticmethod
    def _extract_field(mesg, field):
        try:
            return mesg[field]
        except KeyError:
            return None


def get_kafka_consumer(settings: Mapping, topics: Sequence, fetch_mode: str) -> KafkaConsumer:
    """ Factory method returning a KafkaConsumer object with desired configuration

    :param settings: Mapping object containing project settings
    :param topics: Sequence of topics to subscribe
    :param fetch_mode: str describing fetch mode (from_beginning, last_committed_offset)
    :return: KafkaConsumer object with desired configuration
    """
    if KafkaFetchMode(fetch_mode) is KafkaFetchMode.FROM_BEGINNING:
        group_id = None
    elif KafkaFetchMode(fetch_mode) is KafkaFetchMode.LAST_COMMITED_OFFSET:
        group_id = settings["kafka"].get("group_id", None)
    else:
        raise ValueError(f"{fetch_mode} is not a valid KafkaFetchMode. Valid fetch_modes are: "
                         f"'{KafkaFetchMode.FROM_BEGINNING}' and '{KafkaFetchMode.LAST_COMMITED_OFFSET}'")

    return KafkaConsumer(*topics,
                         group_id=group_id,
                         bootstrap_servers=mapping_util.safe_get_nested(settings, keys=("kafka", "brokers"), default="localhost:9092"),
                         security_protocol=mapping_util.safe_get_nested(settings, keys=("kafka", "security_protocol"), default="PLAINTEXT"),
                         sasl_mechanism=mapping_util.safe_get_nested(settings, keys=("kafka", "sasl_mechanism"), default=None),
                         sasl_plain_username=mapping_util.safe_get_nested(settings, keys=("kafka", "sasl_plain_username"), default=None),
                         sasl_plain_password=mapping_util.safe_get_nested(settings, keys=("kafka", "sasl_plain_password"), default=None),
                         ssl_cafile=mapping_util.safe_get_nested(settings, keys=("kafka", "ssl_cafile"), default=None),
                         auto_offset_reset=mapping_util.safe_get_nested(settings, keys=("kafka", "auto_offset_reset"), default='earliest'),
                         enable_auto_commit=False,
                         consumer_timeout_ms=mapping_util.safe_get_nested(settings, keys=("kafka", "consumer_timeout_ms"), default=1000),
                         heartbeat_interval_ms=mapping_util.safe_get_nested(settings, keys=("kafka", "heartbeat_interval_ms"), default=3000),
                         session_timeout_ms=mapping_util.safe_get_nested(settings, keys=("kafka", "session_timeout_ms"), default=10000),
                         max_poll_interval_ms=mapping_util.safe_get_nested(settings, keys=("kafka", "max_poll_interval_ms"), default=300000))



def as_markdown(blob):
    items = []

    items.append(f"### Dataset:\n\n")

    items.append(f"### Blob:\n\n")
    items.append(f"Blob: _{blob.name}_\n\n")
    items.append(f"Bucket: _{blob.bucket.name}_\n\n")
    items.append(f"Storage class: _{blob.storage_class}_\n\n")
    items.append(f"ID: _{blob.id}_\n\n")
    items.append(f"Size: _{blob.size}_\n\n")
    items.append(f"Updated: _{blob.updated}_\n\n")
    items.append(f"Generation: _{blob.generation}_\n\n")
    items.append(f"Metageneration: _{blob.metageneration}_\n\n")
    items.append(f"Etag: _{blob.etag}_\n\n")
    items.append(f"Owner: _{blob.owner}_\n\n")
    items.append(f"Component count: _{blob.component_count}_\n\n")
    items.append(f"Crc32c: _{blob.crc32c}_\n\n")
    items.append(f"md5_hash: _{blob.md5_hash}_\n\n")
    items.append(f"Cache-control: _{blob.cache_control}_\n\n")
    items.append(f"Content-type: _{blob.content_type}_\n\n")
    items.append(f"Content-disposition: _{blob.content_disposition}_\n\n")
    items.append(f"Content-encoding: _{blob.content_encoding}_\n\n")
    items.append(f"Content-language: _{blob.content_language}_\n\n")

    items.append(f"### Metadata:\n\n")

    if blob.metadata is not None:
        for key, value in blob.metadata.items():
            items.append(f"{key}: _{value}_")
            items.append(f"\n\n")

    return "".join(items)


def as_json(blob):
    return {
        "id": blob.id,
        "size": blob.size,
        "name": blob.name,
        "content_encoding": blob.content_encoding,
        "cache_control": blob.cache_control,
        "metadata": blob.metadata,
        "public_url": blob.public_url,
        "bucket": blob.bucket.name,
        "storage_class": blob.storage_class,
        "updated": blob.updated,
        "generation": blob.generation,
        "metageneration": blob.metageneration,
        "etag": blob.etag,
        "owner": blob.owner,
        "component_count": blob.component_count,
        "crc32c": blob.crc32c,
        "md5_hash": blob.md5_hash,
        "content_type": blob.content_type,
        "content_disposition": blob.content_disposition,
        "content_language": blob.content_language
    }



from pathlib import Path
from os import environ
from collections.abc import Mapping


class EnvStore(Mapping):
    """
    Mapping object for storing and easy accessing of Environment variables
    """

    def __init__(self, path: Path = None, env_setter=None):

        if env_setter is None:
            env_setter = environ

        tmp_dict = {}
        if path is not None:
            with path.open("r") as reader:
                envs = reader.readlines()

                for env in envs:
                    env = env.strip()
                    if env[0] == '#':
                        continue

                    var = env.split('=')
                    tmp_dict[str(var[0])] = str(var[1])

        self._env_store = {**env_setter, **tmp_dict}

    @classmethod
    def safe_create(cls, env_file):
        path = None
        if Path(env_file).exists():
            path = Path(env_file)
        return cls(path)

    def __getitem__(self, key: str):
        if not isinstance(key, str):
            raise ValueError("field should be a str")
        return self._env_store[key]

    def __iter__(self):
        return self._env_store.__iter__()

    def get(self, key):
        if not isinstance(key, str):
            raise ValueError("field should be a str")
        return self._env_store.get(key)

    def keys(self):
        return self._env_store.keys()

    def values(self):
        return self._env_store.values()

    def items(self):
        return self._env_store.items()

    def update(self, envs):
        self._env_store.update(envs)

    def __contains__(self, o: object) -> bool:
        return o in self._env_store

    def __len__(self) -> int:
        return len(self._env_store)



from collections.abc import Mapping


class SettingsStore(Mapping):
    """ Klassen har ansvar for å gjøre settings som eksterne URLer, keys, flagg og andre ressurser tilgjengelige

    """

    def __init__(self, settings_dict: Mapping):
        if not isinstance(settings_dict, Mapping):
            raise TypeError(f"param settings_dict={settings_dict} has to be Mapping")
        self._settings_store = settings_dict

    def __getitem__(self, item):
        if not isinstance(item, str):
            raise ValueError("field should be a str")
        return self._settings_store[item]

    def __iter__(self):
        return self._settings_store.__iter__()

    def __len__(self):
        return len(self._settings_store)

    def __contains__(self, item):
        return item in self._settings_store

    def get(self, key):
        if not isinstance(key, str):
            raise ValueError("field should be a str")
        return self._settings_store.get(key)

    def keys(self):
        return self._settings_store.keys()

    def values(self):
        return self._settings_store.values()

    def items(self):
        return self._settings_store.items()







import os
from dataverk.connectors.storage.bucket_storage_base import BucketStorageBase
from dataverk.connectors.storage.file_storage import FileStorageConnector
from dataverk.connectors.storage.google_storage import GoogleStorageConnector
from collections.abc import Mapping
from enum import Enum
from dataverk.connectors.storage.nais import NaisS3Connector


class StorageType(str, Enum):
    NAIS: str = "nais"
    GCS: str = "gcs"
    LOCAL: str = "local"


def get_storage_connector(
    storage_type: StorageType, settings: Mapping, bucket_name: str = None
) -> BucketStorageBase:
    if storage_type == StorageType.NAIS:
        return NaisS3Connector(bucket_name=bucket_name, s3_endpoint=os.environ["DATAVERK_BUCKET_ENDPOINT"])
    elif storage_type == StorageType.GCS:
        return GoogleStorageConnector(bucket_name=bucket_name, settings=settings)
    elif storage_type == StorageType.LOCAL:
        return FileStorageConnector(settings)
    else:
        raise NotImplementedError(
            f"""Bucket type {storage_type} is not supported.
            Supported types are {[name.value for name in StorageType]}"""
        )



def get_schema(df, path, dataset_name, format, dsv_separator):
    fields = []

    for name, dtype in zip(df.columns, df.dtypes):
        # TODO : Bool and others
        if str(dtype) == 'object':
            dtype = 'string'
        else:
            dtype = 'number'

        fields.append({'name': name, 'description': '', 'type': dtype})

    if format == 'csv':
        mediatype = 'text/csv'
    elif format == 'json':
        mediatype = 'application/json'
    else:
        mediatype = 'text/csv'

    return {
        'name': dataset_name,
        'path': f'{path}/resources/{dataset_name}.{format}',
        'format': format,
        'separator': dsv_separator,
        'mediatype': mediatype,
        'schema': {'fields': fields}
    }





from typing import Sequence, Mapping


def safe_get_nested(mapping: Mapping, keys: Sequence, default):
    for key in keys:
        if key not in mapping:
            return default
        else:
            mapping = mapping[key]
    return mapping


import re


def validate_bucket_name(name):
    ''' Validates that the bucket name consists of lower case letters and numbers, words separated with '-', and
        that it does not start or end with '-'.

    :param name: str: bucket name to be validated
    '''

    valid_name_pattern = "(^[a-z0-9])([a-z0-9\-])+([a-z0-9])$"
    if not re.match(pattern=valid_name_pattern, string=name):
        raise NameError(f"Illegal bucket name ({name}): "
                        "Must be lower case letters or numbers, words separated with '-', "
                        "and cannot start or end with '-')")



from typing import Mapping
from pathlib import Path
from dataverk.exceptions import dataverk_exceptions
from dataverk.connectors.storage.bucket_storage_base import BucketStorageBase
from dataverk.utils import file_functions


class FileStorageConnector(BucketStorageBase):
    """File Storage connector"""

    def __init__(self, settings: Mapping):
        super().__init__()

        try:
            self._path = settings["bucket_storage"]["local"]["path"]
        except KeyError as missing:
            raise dataverk_exceptions.IncompleteSettingsObject(f"{missing}")

    def write(self, data, destination_blob_name: str, fmt: str, **kwargs) -> None:
        """Write resource to file"""
        file_functions.write_file(
            path=f"{Path(self._path).joinpath(destination_blob_name)}.{fmt}",
            content=data,
            compressed=fmt.endswith("gz"),
        )
        self.log.info(
            f"Resource written to file: {self._path}/{destination_blob_name}.{fmt}"
        )

    def read(self, blob_name: str, **kwargs):
        """Read resource from file"""
        data = file_functions.read_file(f"{Path(self._path).joinpath(blob_name)}")
        self.log.info(f"Resource read from file: {self._path}/{blob_name}")
        return data



import requests

from dataverk.abc.base import DataverkBase
from collections.abc import Mapping

from dataverk.exceptions import dataverk_exceptions


class ElasticsearchConnector(DataverkBase):
    """Elasticsearch connector"""

    def __init__(self, settings: Mapping, host="elastic_host"):
        super().__init__()

        self._es_address = self._get_es_address(settings, host)

    def write(self, dp_id: str, doc: Mapping):
        """ Add or update document in es index

        :param dp_id: str: document id
        :param doc: dict: document
        :return:
        """
        try:
            res = requests.post(self._es_address, json=doc, headers={"Content-Type": "application/json"})
            res.raise_for_status()
        except requests.exceptions.HTTPError as err:
            self.log.error(f"Unable to update ES index: {str(err)}""")
            raise requests.exceptions.HTTPError(err)
        except requests.exceptions.RequestException as err:
            self.log.error(f"ES index connection error: {str(err)}")
            raise requests.exceptions.RequestException(err)
        else:
            self.log.info(f"Document {dp_id} successfully written to elastic index.")
            return res

    def _get_es_address(self, settings: Mapping, host):
        try:
            address = settings["index_connections"][host]
        except KeyError as err:
            self.log.warning(f"No ES index specified: {err}")
            raise dataverk_exceptions.IncompleteSettingsObject(f'Connection settings are not available for the host: {host}')
        else:
            return address



from collections.abc import Mapping
from string import Template


class Replacer:

    def __init__(self, value_map: Mapping):
        self._value_map = value_map

    def get_filled_mapping(self, string, reader):
        stemplate = Template(string)
        try:
            filled_string = stemplate.substitute(**self._value_map)
        except KeyError:
            raise KeyError(f"Could not find value for token in settings mapping")
        return reader(filled_string)



import io
import json
import gzip
from collections import Mapping

from dataverk.abc.base import DataverkBase
from dataverk.connectors.storage.bucket_storage_base import BucketStorageBase
from dataverk.connectors.storage.storage_connector_factory import (
    get_storage_connector,
    StorageType,
)


class PackagePublisher(DataverkBase):
    def __init__(
        self, settings_store: Mapping, env_store: Mapping, datapackage_metadata: Mapping
    ):
        super().__init__()
        self._settings_store = settings_store
        self._env_store = env_store
        self._datapackage_metadata = datapackage_metadata

    def publish(self, resources):
        """ - Iterates through all bucket storage connections in the settings.json file and publishes the datapackage
            - Updates ES index with metadata for the datapackage

        :return: None
        """
        bucket_type = self._datapackage_metadata.get("store")

        self.log.info(f"Publishing datapackage {self._datapackage_metadata.get('title')} "
                      f"to bucket {self._datapackage_metadata.get('bucket')}")

        self.upload_to_storage_bucket(
            datapackage_metadata=self._datapackage_metadata,
            conn=get_storage_connector(
                storage_type=StorageType(bucket_type),
                bucket_name=self._datapackage_metadata.get("bucket"),
                settings=self._settings_store
            ),
            datapackage_key_prefix=self._datapackage_key_prefix(
                self._datapackage_metadata.get("id")
            ),
            resources=resources
        )

    @staticmethod
    def upload_to_storage_bucket(
        datapackage_metadata,
        resources,
        conn: BucketStorageBase,
        datapackage_key_prefix: str,
    ) -> None:
        """ Publish data to bucket storage.

        :param resources: datapackage data to be published
        :param datapackage_metadata: metadata associated with the datapackage
        :param conn: BucketStorageConnector object: the connection object for chosen bucket storage.
                     If no bucket storage connector is configured (conn=None) no resources shall be published to bucket storage
        :param datapackage_key_prefix: str: prefix for datapackage key
        :return: None
        """
        conn.write(
            json.dumps(datapackage_metadata),
            datapackage_key_prefix + "datapackage",
            "json", metadata=datapackage_metadata
        )
        for filename, item in resources.items():
            df = item['df']
            sep = item['dsv_separator']
            fmt = item["format"]
            compressed = item["compressed"]

            data_buff = io.StringIO()
            df.to_csv(data_buff, sep=sep, index=False)

            if compressed:
                compressed_data = PackagePublisher._compress_content(data_buff)
                PackagePublisher._upload(conn=conn,
                                         data=compressed_data,
                                         destination_blob_name=f"{datapackage_key_prefix}resources/{filename}",
                                         metadata=datapackage_metadata,
                                         fmt=f"{fmt}.gz")
            else:
                PackagePublisher._upload(conn=conn,
                                         data=data_buff.getvalue(),
                                         destination_blob_name=f"{datapackage_key_prefix}resources/{filename}",
                                         metadata=datapackage_metadata,
                                         fmt=fmt)

    @staticmethod
    def _compress_content(data_buff):
        gz_buff = io.BytesIO()
        with gzip.GzipFile(fileobj=gz_buff, mode='w') as zipped_f:
            zipped_f.write(bytes(data_buff.getvalue(), encoding="utf-8"))
        return gz_buff.getvalue()

    @staticmethod
    def _upload(conn, data, destination_blob_name, metadata, fmt):
        conn.write(data=data,
                   destination_blob_name=destination_blob_name,
                   metadata=metadata,
                   fmt=fmt)

    @staticmethod
    def _datapackage_key_prefix(base: str):
        return base + "/"



import requests
from dataverk.connectors.storage.bucket_storage_base import BucketStorageBase


class NaisS3Connector(BucketStorageBase):
    """
    Connector for NAIS S3 API
    """

    def __init__(self, bucket_name: str, s3_endpoint: str):
        super().__init__()
        self._bucket_name = bucket_name
        self._s3_api_url = s3_endpoint

    def write(self, data, destination_blob_name: str, fmt: str = "csv", **kwargs):
        try:
            res = requests.put(
                url=f"{self._s3_api_url}/{self._bucket_name}/{destination_blob_name}.{fmt}",
                data=data,
            )
            res.raise_for_status()
        except requests.exceptions.HTTPError as err:
            self.log.error(
                f"Unable to write object {destination_blob_name} to bucket {self._bucket_name}: {str(err)}"
            )
            raise requests.exceptions.HTTPError(err)
        except requests.exceptions.RequestException as err:
            self.log.error(f"Connection error {self._s3_api_url}: {str(err)}")
            raise requests.exceptions.RequestException(err)
        else:
            self.log.info(
                f"Object {destination_blob_name} written to bucket {self._bucket_name}"
            )

    def read(self, blob_name: str, **kwargs):
        try:
            res = requests.get(
                url=f"{self._s3_api_url}/{self._bucket_name}/{blob_name}"
            )
            res.raise_for_status()
        except requests.exceptions.HTTPError as err:
            self.log.error(
                f"Unable to read object {blob_name} from bucket {self._bucket_name}: {str(err)}"
            )
            raise requests.exceptions.HTTPError(err)
        except requests.exceptions.RequestException as err:
            self.log.error(f"Connection error {self._s3_api_url}: {str(err)}")
            raise requests.exceptions.RequestException(err)
        else:
            self.log.info(
                f"Object {blob_name} successfully read from bucket {self._bucket_name}"
            )
            return res.text



import time
import pandas as pd

from typing import Mapping
from sqlalchemy import engine
from sqlalchemy.exc import SQLAlchemyError
from dataverk.abc.base import DataverkBase


class DBBaseConnector(DataverkBase):
    def __init__(self, settings_store: Mapping, source: str):
        super().__init__()
        self.settings = settings_store
        self.source = source

    def __enter__(self):
        self._engine = self._create_engine()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._engine.dispose()

    def get_pandas_df(
        self, query: str, verbose_output: bool = False, *args, **kwargs
    ) -> pd.DataFrame:
        start_time = time.time()
        self.log.info(f"Reading from database: {self.source}")

        try:
            df = pd.read_sql(query, self._engine, *args, **kwargs)
        except SQLAlchemyError as error:
            self.log.error(f"{error}")
            raise SQLAlchemyError(f"{error}")

        end_time = time.time()
        self.log.info(f"{len(df)} records returned in {end_time - start_time} seconds.")
        if verbose_output:
            self.log.info(f"Query: {query}")

        return df

    def persist_pandas_df(self, table: str, df: pd.DataFrame, *args, **kwargs) -> None:
        start_time = time.time()
        self.log.info(
            f"Persisting {len(df)} records to table: {table} in database: {self.source}"
        )

        try:
            df.to_sql(name=table, con=self._engine, *args, **kwargs)
        except SQLAlchemyError as error:
            self.log.error(f"{error}")
            raise SQLAlchemyError(f"{error}")

        end_time = time.time()
        self.log.info(
            f"Persisted {len(df)} records to table {table} in {end_time - start_time} seconds"
        )

    def _create_engine(self) -> engine.Engine:
        db = self._connection_string()
        return engine.create_engine(db)

    def _connection_string(self) -> str:
        return self.settings["db_connection_strings"][self.source]



import requests
import json
from pyjstat import pyjstat
from collections import OrderedDict
import pandas as pd
import ast
from dataverk.abc.base import DataverkBase


class JSONStatConnector(DataverkBase):
    """JSONStat based connections
    
    """
    
    def __init__(self):
        super().__init__()
     
    def get_pandas_df(self, url, params=None, table_format='json'):
        """
        Get Pandas dataframe
        """

        self.log(str(url))

        if params == None:
            params = json.dumps(self._full_json(url))
           
        response=requests.post(url,params).content
        response=response.decode('utf-8')
        df = pyjstat.from_json_stat(json.loads(response))[0]
        return df

    def _get_table(self, url, table_format = 'json'):

        if table_format == 'json':
            response = requests.get(url)
            df = pyjstat.from_json_stat(response.json(object_pairs_hook=OrderedDict))[0]
        elif table_format == 'csv':
            df = pd.read_csv(url)
        else:
            print("""table_format param must be either 'json' or 'csv'""")
            df = None
        return df

    def _full_json(self,url):
        variables = self._get_variables(url)
        nvars = len(variables)
        var_list = list(range(nvars))
        
        query_element = {}
        
        for x in var_list:
            query_element[x] ='{{"code": "{code}", "selection": {{"filter": "item", "values": {values} }}}}'.format(
                    code = variables[x]['code'], 
                    values = variables[x]['values']) 
            query_element[x] = query_element[x].replace("\'", '"')
        all_elements = str(list(query_element.values()))
        all_elements = all_elements.replace("\'", "")
        
        query = '{{"query": {all_elements} , "response": {{"format": "json-stat" }}}}'.format(all_elements = all_elements)

        query = ast.literal_eval(query)
        
        return query

    def _get_variables(self,url):
        df = pd.read_json(url)
        variables = [values for values in df.iloc[:,1]]
        return variables




import json
from dataverk.context import EnvStore
from dataverk.context import values_importer
from dataverk.context.replacer import Replacer
from dataverk.utils import file_functions


class DataverkContext:

    def __init__(self, env_store: EnvStore, resource_path: str=".", auth_token: str=None):
        self._resource_path = resource_path
        self._http_headers = self._set_http_headers(auth_token)
        self._env_store = env_store
        self._settings_store = self._load_settings()
        self._load_and_apply_secrets()

    @property
    def settings(self):
        return self._settings_store

    def _load_settings(self):
        return json.loads(file_functions.get_package_resource("settings.json", self._resource_path, self._http_headers))

    def _load_and_apply_secrets(self):
        self._settings_store = self._apply_secrets(self._env_store, self._settings_store)

    def get_sql_query(self, sql: str):
        return file_functions.get_package_resource(sql, self._resource_path, self._http_headers)

    @staticmethod
    def _set_http_headers(auth_token: str):
        if auth_token is not None:
            return {"Authorization": f"token {auth_token}"}
        else:
            return None

    @staticmethod
    def _apply_secrets(env_store, settings):
        importer = values_importer.get_secrets_importer(settings, env_store)
        replacer = Replacer(importer.import_values())
        return replacer.get_filled_mapping(json.dumps(settings), json.loads)



import time
import pandas as pd
from dataverk.exceptions import dataverk_exceptions

from sqlalchemy.exc import SQLAlchemyError, OperationalError
from collections.abc import Mapping
from dataverk.connectors.databases.base import DBBaseConnector
from dataverk.connectors.databases.utils.error_strategies import (
    PostgresOperationalErrorStrategy,
    ErrorStrategy,
)


class PostgresConnector(DBBaseConnector):
    def __init__(
        self,
        settings_store: Mapping,
        source: str,
        error_strategy: ErrorStrategy = PostgresOperationalErrorStrategy(),
    ):
        super().__init__(settings_store, source)
        self.error_strategy = error_strategy

    def get_pandas_df(
        self, query: str, verbose_output: bool = False, *args, **kwargs
    ) -> pd.DataFrame:
        start_time = time.time()
        self.log.info(f"Reading from PostgreSQL database: {self.source}")

        try:
            df = pd.read_sql_query(query, self._engine, *args, **kwargs)
        except OperationalError:
            self.error_strategy.handle_error(self)
            df = pd.read_sql_query(query, self._engine, *args, **kwargs)
        except SQLAlchemyError as error:
            self.log.error(f"{error}")
            raise SQLAlchemyError(f"{error}")

        end_time = time.time()
        self.log.info(f"{len(df)} records returned in {end_time - start_time} seconds.")
        if verbose_output:
            self.log(f"Query: {query}")

        return df

    def persist_pandas_df(self, table: str, df: pd.DataFrame, *args, **kwargs) -> None:
        self.log.info(
            f"Persisting {len(df)} records to table: {table} in PostgreSQL database: {self.source}"
        )
        start_time = time.time()

        try:
            self._set_role()
            df.to_sql(table, self._engine, *args, **kwargs)
        except OperationalError:
            self.error_strategy.handle_error(self)
            self._set_role()
            df.to_sql(table, self._engine, *args, **kwargs)
        except SQLAlchemyError as error:
            self.log.error(f"{error}")
            raise SQLAlchemyError(f"{error}")

        end_time = time.time()
        self.log.info(
            f"Persisted {len(df)} records to table {table} in {end_time - start_time} seconds"
        )

    def _get_role_name(self) -> str:
        try:
            vault_path = self.settings["db_vault_path"][self.source]
        except KeyError as missing:
            raise dataverk_exceptions.IncompleteSettingsObject(f"{missing}")
        else:
            return f"{vault_path.split('/')[-1]}"

    def _set_role(self) -> None:
        try:
            query = f"SET ROLE '{self._get_role_name()}'; COMMIT;"
        except dataverk_exceptions.IncompleteSettingsObject as err:
            self.log.error(f"Unable to set role: {err}")
            raise dataverk_exceptions.IncompleteSettingsObject(f"{err}")
        else:
            self._engine.execute(query)



import copy
import pandas as pd
import datetime
import uuid
import hashlib
import re
from urllib3.util import url
from os import environ

from data_catalog_dcat_validator.models.dataset import DatasetModel
from dataverk.exceptions.dataverk_exceptions import EnvironmentVariableNotSet
from dataverk.utils import validators, file_functions
from collections.abc import Mapping, Sequence
from dataverk.connectors.storage.storage_connector_factory import StorageType


class Datapackage:
    """
    Understands packaging of data resources and views on those resources for publication
    """

    def __init__(self, metadata: Mapping):
        self._resources = {}
        self.views = []
        self._validate_metadata(metadata)
        self._datapackage_metadata = self._create_datapackage(dict(metadata))

    def _create_datapackage(self, metadata):
        today = datetime.date.today().strftime('%Y-%m-%d')

        try:
            bucket = metadata['bucket']
        except KeyError:
            raise AttributeError(f"bucket is required to be set in datapackage metadata")
        else:
            validators.validate_bucket_name(bucket)

        try:
            metadata['title']
        except KeyError:
            raise AttributeError(f"title is required to be set in datapackage metadata")

        # set defaults for store and repo when not specified
        metadata['store'] = metadata.get('store', StorageType.LOCAL)
        metadata['repo'] = metadata.get('repo', metadata.get('github-repo', ''))

        try:
            dp_id = metadata['id']
        except KeyError:
            dp_id = self._generate_id(metadata)

        metadata['id'] = dp_id
        path, store_path = self._generate_paths(metadata)

        metadata['store_path'] = store_path
        metadata['path'] = path
        metadata['updated'] = today
        metadata['version'] = "0.0.1"
        metadata["views"] = []
        metadata["resources"] = []
        metadata["datasets"] = {}
        return metadata

    @staticmethod
    def _validate_metadata(metadata: Mapping):
        validator = DatasetModel(metadata)
        validator.validate()
        validator.error_report()

    @property
    def datapackage_metadata(self):
        return copy.deepcopy(self._datapackage_metadata)

    @property
    def resources(self):
        return self._resources

    @property
    def dp_id(self):
        return self._datapackage_metadata.get("id")

    @property
    def project(self):
        return self._datapackage_metadata.get("project")

    @property
    def path(self):
        return self._datapackage_metadata.get("path")

    @property
    def url(self):
        return self._datapackage_metadata.get("url")

    def _get_schema(self, df, path, resource_name, resource_description, format, compress, dsv_separator, spec):
        fields = []

        for name, dtype in zip(df.columns, df.dtypes):
            if str(dtype) == 'object':
                dtype = 'string'
            else:
                dtype = 'number'

            description = ''
            if spec and spec.get('fields'):
                spec_fields = spec['fields']
                for field in spec_fields:
                    if field['name'] == name:
                        description = field['description']

            fields.append({'name': name, 'description': description, 'type': dtype})

        mediatype = self._media_type(format)

        return {
            'name': resource_name,
            'description': resource_description,
            'path': self._resource_path(path, resource_name, format, compress),
            'format': format,
            'dsv_separator': dsv_separator,
            'compressed': compress,
            'mediatype': mediatype,
            'schema': {'fields': fields},
            'spec': spec
        }

    @staticmethod
    def _media_type(fmt):
        if fmt == 'csv':
            return 'text/csv'
        elif fmt == 'json':
            return 'application/json'
        else:
            return 'text/csv'

    @staticmethod
    def _resource_path(path, resource_name, fmt, compress):
        if compress:
            return f'{path}/resources/{resource_name}.{fmt}.gz'
        else:
            return f'{path}/resources/{resource_name}.{fmt}'

    def add_resource(self, df: pd.DataFrame, resource_name: str, resource_description: str="",
                     format="csv", compress: bool=True, dsv_separator=";", spec: Mapping=None):
        """
        Adds a provided DataFrame as a resource in the Datapackage object with provided name and description.

        :param df: DataFrame to add as resource
        :param resource_name: Name of the resource
        :param resource_description: Description of the resource
        :param format: file format of resource
        :param compress: boolean value indicating whether to compress resource before storage
        :param dsv_separator: field separator
        :param spec: resource specification
        :return: None
        """
        self._verify_add_resource_input_types(df, resource_name, resource_description)
        resource_name = file_functions.remove_whitespace(resource_name)
        self.resources[resource_name] = self._get_schema(df=df, path=self.path, resource_name=resource_name,
                                                         resource_description=resource_description, format=format,
                                                         compress=compress, dsv_separator=dsv_separator, spec=spec)
        self.resources[resource_name]['df'] = df
        self._datapackage_metadata["datasets"][resource_name] = resource_description
        self._datapackage_metadata['resources'].append(self._get_schema(df=df, path=self.path,
                                                                        resource_name=resource_name,
                                                                        resource_description=resource_description,
                                                                        format=format, compress=compress,
                                                                        dsv_separator=dsv_separator, spec=spec))

    def add_remote_resource(self, resource_url: str, resource_description: str=""):
        resource_name, resource_fmt = self._resource_name_and_type_from_url(resource_url)
        self._datapackage_metadata['datasets'][resource_name] = resource_description
        self._datapackage_metadata['resources'].append({
            'name': resource_name,
            'description': resource_description,
            'path': resource_url,
            'format': resource_fmt
        })

    @staticmethod
    def _resource_name_and_type_from_url(resource_url):
        parsed_url = url.parse_url(resource_url)

        if not parsed_url.scheme == "https" and not parsed_url.scheme == "http":
            raise ValueError(f"Remote resource needs to be a web address, scheme is {parsed_url.scheme}")

        resource = parsed_url.path.split('/')[-1]
        resource_name_and_format = resource.split('.', 1)
        return resource_name_and_format[0], resource_name_and_format[1]

    @staticmethod
    def _verify_add_resource_input_types(df, dataset_name, dataset_description):
        if not isinstance(df, pd.DataFrame):
            raise TypeError(f'df must be of type pandas.Dataframe()')
        if not isinstance(dataset_name, str):
            raise TypeError(f'dataset_name must be of type string')
        if not isinstance(dataset_description, str):
            raise TypeError(f'dataset_description must be of type string')

    def add_view(self, name: str, resources: Sequence, title: str="", description: str="", attribution: str="", spec_type: str="simple",
                 spec: Mapping=None, type: str="", group: str="", series: Sequence=list(), row_limit: int=500, metadata: Mapping=None):
        """
        Adds a view to the Datapackage object. A view is a specification of a visualisation the datapackage provides.

        :param name: View name
        :param resources: resource the view is for
        :param title: Title to be presented in the visualisation
        :param description: Description of the view
        :param attribution:
        :param spec_type: Spec type eg. (matplotlib, vega-lite)
        :param spec:
        :param type:
        :param group:
        :param series:
        :param row_limit:
        :param metadata:
        :return: None
        """
        if spec is None:
            spec = {"type": type,
                    "group": group,
                    "series": series}

        view = {'name': file_functions.remove_whitespace(name),
                'title': title,
                'description': description,
                'attribution': attribution,
                'resources': resources,
                'specType': spec_type,
                'spec': spec,
                'transform': {
                    "limit": row_limit
                },
                'metadata': metadata
                }

        self._datapackage_metadata["views"].append(view)

    @staticmethod
    def _generate_id(metadata):
        author = metadata.get("author", None)
        title = metadata.get("title", None)
        bucket = metadata.get("bucket", None)

        id_string = '-'.join(filter(None, (bucket, author, title)))
        if id_string:
            hash_object = hashlib.md5(id_string.encode())
            dp_id = hash_object.hexdigest()
            return re.sub('[^0-9a-z]+', '-', dp_id.lower())
        else:
            return uuid.uuid4()

    @staticmethod
    def _generate_paths(metadata):
        store = metadata['store']
        repo = metadata['repo']
        bucket = metadata['bucket']
        dp_id = metadata['id']

        if StorageType(store) is StorageType.NAIS:
            path, store_path = Datapackage._nais_specific_paths(bucket, dp_id)
        elif StorageType(store) is StorageType.GCS:
            path = f'https://storage.googleapis.com/{bucket}/{dp_id}'
            store_path = f'gs://{bucket}/{dp_id}'
        else: #default is local storage
            path = f'https://raw.githubusercontent.com/{repo}/master/{bucket}/packages/{dp_id}'
            store_path = f'{bucket}/{dp_id}'

        return path, store_path

    @staticmethod
    def _nais_specific_paths(bucket, dp_id):
        try:
            api_endpoint = environ["DATAVERK_API_ENDPOINT"]
        except KeyError as missing_env:
            raise EnvironmentVariableNotSet(missing_env)
        else:
            path = f'{api_endpoint}/{bucket}/{dp_id}'

        try:
            bucket_endpoint = environ["DATAVERK_BUCKET_ENDPOINT"]
        except KeyError as missing_env:
            raise EnvironmentVariableNotSet(missing_env)
        else:
            store_path = f'{bucket_endpoint}/{bucket}/{dp_id}'

        return path, store_path



import json
from collections import Sequence, Mapping
from os import environ

import pandas as pd
import requests

from dataverk.exceptions.dataverk_exceptions import EnvironmentVariableNotSet


def anonymize_replace(df: pd.DataFrame, eval_column, anonymize_columns=None, evaluator=lambda x: x < 4,
                      replace_by="*", anonymize_eval=True) -> pd.DataFrame:
    _check_valid_anonymization(anonymize_columns, anonymize_eval)
    columns = _set_columns_to_anonymize(df, eval_column, anonymize_columns, anonymize_eval)

    columns, replace_by = _set_replace_by(df, columns, replace_by)

    to_anonymize = df.copy()
    return _replace_value(to_anonymize, eval_column, columns, evaluator, replace_by)


def _check_valid_anonymization(anonymize_columns, anonymize_eval):
    if anonymize_columns is None and not anonymize_eval:
        raise ValueError("df will not be anonymized, no additional columns are given and anonymize_eval is set to False")


def _set_columns_to_anonymize(df, eval_column, anonymize_columns, anonymize_eval):
    columns = _check_anonymize_columns_input_type(anonymize_columns)

    if anonymize_eval and eval_column not in columns:
        columns += [eval_column]

    if not anonymize_eval and eval_column in columns:
        columns = [col for col in columns if col != eval_column]

    _check_column_names(df, columns)
    return columns


def _check_anonymize_columns_input_type(anonymize_columns):
    if anonymize_columns is None:
        anonymize_columns = []

    elif not isinstance(anonymize_columns, Sequence) or isinstance(anonymize_columns, str):
        anonymize_columns = [anonymize_columns]

    return anonymize_columns


def _check_column_names(df, columns):
    for column in columns:
        if column not in df.columns:
            raise AttributeError(f"'{column}' is not a column in df")


def _set_replace_by(df, columns, replace_by):
    column_order, replace_by_order = columns, replace_by

    if isinstance(replace_by, Mapping):
        replace_in_columns = list(replace_by.keys())
        replace_by_values = list(replace_by.values())

        _check_column_names(df, replace_in_columns)
        column_order, replace_by_order = replace_in_columns, replace_by_values

    return column_order, replace_by_order


def _replace_value(df, eval_column, columns, evaluator, replace_by):
    df.loc[df[df[eval_column].apply(evaluator)].index, columns] = replace_by
    return df


def name_replace(df, columns) -> pd.DataFrame:
    """ Replaces names in columns

    :param df: pandas DataFrame
    :param columns: list of columns to apply name replacement
    :return: pandas DataFrame
    """
    try:
        url = environ["DATAVERK_NAME_REPLACE_API"]
    except KeyError as missing_env:
        raise EnvironmentVariableNotSet(missing_env)

    to_anonymize = df.copy()
    for column in columns:
        res = requests.post(url, data={'values': json.dumps(to_anonymize[column].tolist())})
        filtered_list = json.loads(res.text)['result']
        to_anonymize[column] = pd.Series(filtered_list)
    return to_anonymize



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from pathlib import Path

# Gjør README.md om til den lange beskrivelsen på PiPy
with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    install_requires = f.read().strip().split('\n')

setup(
    name='dataverk',
    version='0.2.24',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    python_requires='>=3.6',
    install_requires=install_requires,
    package_data={
        'dataverk': ['VERSION']
    },
    # metadata to display on PyPI
    author="NAV IKT",
    author_email="paul.bencze@nav.no",
    description="Serverless ETL framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    keywords="datapackage serverless etl jupyter open-data",
    url="https://github.com/navikt",
    project_urls={
        "Bug Tracker": "https://github.com/navikt",
        "Documentation": "https://github.com/navikt",
        "Source Code": "https://github.com/navikt",
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
)



from dataverk_vault import api as vault_api

from abc import ABC, abstractmethod
from urllib3.util import url
from dataverk.connectors.databases.base import DBBaseConnector
from dataverk.exceptions import dataverk_exceptions


class ErrorStrategy(ABC):
    @staticmethod
    @abstractmethod
    def handle_error(connector: DBBaseConnector):
        raise NotImplementedError()


class PostgresOperationalErrorStrategy(ErrorStrategy):

    @staticmethod
    def handle_error(connector: DBBaseConnector):
        connector._engine.dispose()

        try:
            vault_path = connector.settings["db_vault_path"][connector.source]
        except KeyError as err:
            connector.log.error(
                f"No vault path specified in settings, unable to update credentials: {err}"
            )
            raise dataverk_exceptions.IncompleteSettingsObject(f"{err}")
        else:
            PostgresOperationalErrorStrategy._update_credentials(connector, vault_path)
            connector._engine = connector._create_engine()

    @staticmethod
    def _update_credentials(connector: DBBaseConnector, vault_path: str):
        connector.log.warning(f"Updating db credentials")
        conn_string = connector._connection_string()
        parsed_url = url.parse_url(conn_string)
        new_credentials = vault_api.get_database_creds(
            vault_path
        )
        connector.settings["db_connection_strings"][
            connector.source
        ] = conn_string.replace(parsed_url.auth, new_credentials)



from dataverk.connectors.jsonstat import JSONStatConnector
from dataverk.connectors.storage.google_storage import GoogleStorageConnector
from dataverk.connectors.databases.db2 import Db2Connector
from dataverk.connectors.databases.oracle import OracleConnector
from dataverk.connectors.databases.postgres import PostgresConnector
from dataverk.connectors.databases.sqlite import SqliteConnector
from dataverk.connectors.kafka import KafkaConnector
from dataverk.connectors.storage.nais import NaisS3Connector


__all__ = [
            'JSONStatConnector',
            'NaisS3Connector',
            'GoogleStorageConnector',
            'OracleConnector',
            'KafkaConnector',
            'PostgresConnector',
            'Db2Connector',
            'SqliteConnector'
           ]



import os
from abc import ABC
from dataverk.utils.logger import Logger


class DataverkBase(ABC):
    """
    Base class for all dataverk classes
    """
    def __init__(self):
        self._logger = Logger(user=self._get_user(),
                              connector_name=self._get_class_name())

    @property
    def log(self):
        return self._logger.log

    @staticmethod
    def _get_user():
        try:
            import pwd
            return pwd.getpwuid(os.getuid()).pw_name
        except ImportError:
            return os.getlogin()

    def _get_class_name(self):
        return self.__class__.__name__



from pathlib import Path
from collections.abc import Mapping
from abc import ABC, abstractmethod
from urllib import parse
from urllib.parse import ParseResult
import dataverk_vault.api as vault_api


class ValuesImporter(ABC):

    @abstractmethod
    def import_values(self) -> Mapping:
        pass


class FileValuesImporter(ValuesImporter):

    def __init__(self, resource):
        super().__init__()
        self._resource_path = Path(resource)

    def import_values(self) -> Mapping:
        token_value_map = {}
        for value in self._resource_path.iterdir():
            token_value_map.update(self._import_value(value))
        return token_value_map

    @staticmethod
    def _import_value(value):
        token_value_map = {}
        try:
            f = value.open("r")
        except OSError:
            raise Warning(f"Could not open value={value}")
        else:
            token_value_map[value.name] = f.read()
            f.close()
        return token_value_map


class APIValuesImporter(ValuesImporter):

    def __init__(self):
        super().__init__()

    def import_values(self) -> Mapping:
        return vault_api.read_secrets()


class NullValuesImporter(ValuesImporter):

    def import_values(self):
        return {"": ""}


def get_secrets_importer(settings: Mapping, env_store: Mapping) -> ValuesImporter:
    if env_store.get("DATAVERK_SECRETS_FROM_FILES") is not None:
        return FileValuesImporter(resource=settings["secret_path"])
    elif env_store.get("DATAVERK_SECRETS_FROM_API") is not None:
        return APIValuesImporter()
    else:
        return NullValuesImporter()


def _parse_url(url: str) -> ParseResult:
    return parse.urlparse(url)
